package com.jakubu9333.grammarManager.api;

import com.fasterxml.jackson.annotation.JsonView;
import com.jakubu9333.grammarManager.model.FileHolder;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.View;
import com.jakubu9333.grammarManager.service.FileService;
import com.jakubu9333.grammarManager.service.GrammarService;
import com.jakubu9333.grammarManager.service.RuleService;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.CrossOrigin;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.server.ResponseStatusException;

import javax.servlet.http.HttpServletResponse;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

/**
 * @author Jakub Uhlarik
 */
@RestController
@CrossOrigin(origins = "http://localhost:3000")
@RequestMapping(path = "api/grammars")
public class GrammarController {
    public final RuleService ruleService;

    private final GrammarService grammarService;

    private final FileService fileService;

    @Autowired
    public GrammarController(RuleService ruleService, GrammarService grammarService, FileService fileService) {
        this.ruleService = ruleService;
        this.grammarService = grammarService;
        this.fileService = fileService;
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @GetMapping("all")
    public ResponseEntity<List<Grammar>> getAllGrammars() {
        return new ResponseEntity<>(grammarService.getGrammars(), HttpStatus.OK);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @GetMapping("{id}")
    public ResponseEntity<Grammar> getGrammar(@PathVariable Long id) {
        return new ResponseEntity<>(grammarService.getGrammarById(id), HttpStatus.OK);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("")
    public ResponseEntity<Grammar> addGrammar(@RequestBody Grammar grammar) {
        return new ResponseEntity<>(grammarService.addGrammar(grammar), HttpStatus.CREATED);
    }

    @DeleteMapping("/all")
    public ResponseEntity<?> deleteAllGrammar() {
        grammarService.deleteAll();
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @DeleteMapping("/{id}")
    public ResponseEntity<?> deleteGrammar(@PathVariable Long id) {
        grammarService.deleteById(id);
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("{id}/duplicate")
    public ResponseEntity<Grammar> duplicateGrammar(@PathVariable Long id) {
        return new ResponseEntity<>(grammarService.duplicateGrammar(id), HttpStatus.CREATED);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("/file/{fileFormat}")
    public ResponseEntity<Grammar> grammarFromFile(@RequestBody String file, @PathVariable String fileFormat, @RequestParam(defaultValue = "true") boolean save) throws IOException {
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> xml = false;

            case "xml" -> xml = true;

            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Grammar grammar = fileService.addGrammarFromFile(file, xml, null, false, save);
        return new ResponseEntity<>(grammar, HttpStatus.CREATED);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("{id}/file/{fileFormat}")
    public ResponseEntity<Grammar> grammarFromFile(@RequestBody String file, @PathVariable String fileFormat, @PathVariable Long id, @RequestParam(defaultValue = "false") boolean newGrammar, @RequestParam(defaultValue = "true") boolean save) throws IOException {
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> xml = false;

            case "xml" -> xml = true;

            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Grammar grammar = fileService.addGrammarFromFile(file, xml, id, newGrammar, save);
        if (grammar == null) {
            return new ResponseEntity<>(HttpStatus.NO_CONTENT);
        }
        return new ResponseEntity<>(grammar, HttpStatus.CREATED);
    }


    @JsonView(View.RuleWithoutGrammar.class)
    @GetMapping("/rootRule")
    public ResponseEntity<List<Grammar>> findGrammarByRootRule(@RequestParam String root) {
        return new ResponseEntity<>(grammarService.getGrammarByRootRule(root), HttpStatus.OK);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @GetMapping("/uri/")
    public ResponseEntity<List<Grammar>> findGrammarByUriStart(@RequestParam String uri) {
        return new ResponseEntity<>(grammarService.getGrammarsByUriStart(uri), HttpStatus.OK);
    }


    @GetMapping("{id}/file/new/{fileFormat}")
    public ResponseEntity<InputStreamResource> fileFromGrammar(@PathVariable String fileFormat, @PathVariable Long id) throws IOException {
        MediaType typeOfFile;
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> {
                typeOfFile = MediaType.TEXT_PLAIN;
                xml = false;
            }
            case "xml" -> {
                typeOfFile = MediaType.APPLICATION_XML;
                xml = true;
            }
            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Pair<FileHolder, Boolean> result = fileService.newFileFromGrammar(id, xml);
        InputStream is = new FileInputStream("./grammars/" + result.getValue0().getFilePath());
        return ResponseEntity.status(result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK).contentType(typeOfFile).body(new InputStreamResource(is));
    }

    @GetMapping("{id}/file/new/{fileFormat}/download")
    public ResponseEntity<InputStreamResource> fileFromGrammarDownload(@PathVariable String fileFormat, @PathVariable Long id, HttpServletResponse response) throws IOException {
        MediaType typeOfFile;
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> {
                typeOfFile = MediaType.TEXT_PLAIN;
                xml = false;
            }
            case "xml" -> {
                typeOfFile = MediaType.APPLICATION_XML;
                xml = true;
            }
            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        FileHolder fileHolder = fileService.getGeneratedFile(id, xml);
        InputStream is = new FileInputStream("./grammars/" + fileHolder.getFilePath());
        response.setHeader("Content-Disposition", "attachment; filename=" + fileHolder.getFilePath());
        return ResponseEntity.ok().contentType(typeOfFile).body(new InputStreamResource(is));
    }

    @GetMapping("file/{id}")
    public ResponseEntity<InputStreamResource> getFile(@PathVariable Long id) throws IOException {
        MediaType typeOfFile;
        FileHolder fileHolder = fileService.getFileHolderById(id);
        if (fileHolder.isXml()) {
            typeOfFile = MediaType.APPLICATION_XML;
        } else {
            typeOfFile = MediaType.TEXT_PLAIN;
        }
        InputStream is = new FileInputStream("./grammars/" + fileHolder.getFilePath());
        return ResponseEntity.ok().contentType(typeOfFile).body(new InputStreamResource(is));
    }

    @GetMapping("/file/{id}/download")
    public ResponseEntity<InputStreamResource> getFileDownload(@PathVariable Long id, HttpServletResponse response) throws IOException {
        MediaType typeOfFile;
        FileHolder fileHolder = fileService.getFileHolderById(id);
        if (fileHolder.isXml()) {
            typeOfFile = MediaType.APPLICATION_XML;
        } else {
            typeOfFile = MediaType.TEXT_PLAIN;
        }
        InputStream is = new FileInputStream("./grammars/" + fileHolder.getFilePath());
        response.setHeader("Content-Disposition", "attachment; filename=" + fileHolder.getFilePath());
        return ResponseEntity.ok().contentType(typeOfFile).body(new InputStreamResource(is));
    }


    @JsonView(View.RuleWithoutGrammar.class)
    @PutMapping("{id}/uri")
    public ResponseEntity<Grammar> changeGrammarUri(@PathVariable Long id, @RequestBody Grammar grammar) {
        return new ResponseEntity<>(grammarService.changeUri(id, grammar), HttpStatus.OK);
    }


}
